# -*- coding: utf-8 -*-
"""Deployment scripts."""
import os
import random
import sys

import click
import envoy
import yaml

import git
import heroku

# # Config file can be supplied, but need a default, which is in the same dir
DEFAULT_CONFIG = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'deploy.yaml'
)


class Configuration():

    """Encapsulate deployment configuration, with helper methods."""

    def __init__(self, environments, commands):
        """Initialise with environments list and commands dict."""
        self.environments = environments
        self.commands = commands

    @classmethod
    def load(cls, filename):
        """Create new object from file."""
        with open(filename, 'r') as f:
            config = yaml.load(f)
        return Configuration(
            environments=config['environments'],
            commands=config['commands']
        )

    def get_command(self, command):
        """Return command from the configuration."""
        return self.commands[command]

    def get_environment(self, environment):
        """Return a single environment config as a dict."""
        return [e for e in self.environments if e['label'] == environment][0]

    def get_environment_setting(self, environment, setting, default=None):
        """Return a single environment setting."""
        return self.get_environment(environment).get(setting, default)


def prompt_for_pin(prompt, exit_on_failure=True):
    """Prompt user to input random number before continuing.

    Args:
        pre_prompt: a message to print out before the prompt.

    Kwargs:
        exit_on_failure: if True (default) then immediately call sys.exit()
            if the user does not enter the correct PIN; if False, then
            return False to the calling procedure instead.

    """
    p = str(random.randint(0, 10 ** 6 - 1))
    if prompt is not None:
        click.echo(prompt)
    pin = raw_input(
        "Please confirm your intention to continue by typing in '%s' "
        "at the prompt: " % p
    )
    if pin == p:
        return True
    else:
        if exit_on_failure:
            print("PIN incorrect.")
            sys.exit(0)
        else:
            return False


def get_release_note(changelog, filename="RELEASE_NOTE"):
    """Invoke git editor to create a new release note."""
    editor = git.get_editor()
    if editor == '':
        print "No editor specified."
        return changelog
    else:
        with open(filename, 'w') as f:
            f.write(changelog)
        envoy.run('%s %s' % (editor, filename))
        with open(filename, 'r') as f:
            release_note = f.read()
        os.remove(filename)
        return release_note


def get_remote_commit_hash(application):
    """Get commit hash of Heroku application."""
    return str(heroku.get_latest_release(application)[1])


def get_local_commit_hash(branch):
    """Get the hash of the latest commit on a given branch."""
    return git._do_git_cmd("rev-parse %s" % branch)[:7]


def prompt_for_action(question, default):
    """Return confirmation from user for an action."""
    if default is True:
        # if the default is True, the a 'y' or nothing is good
        action = raw_input(question+' [Y/n]: ')
        return action == '' or action.lower().startswith('y')
    else:
        # if the default is False, then only a 'y' will do
        action = raw_input(question+' [y/N]: ')
        return action.lower().startswith('y')


def split_print_lines(text, line_format='  * %s'):
    """Split a block of text and print out as lines."""
    for line in text.lstrip().rstrip().split('\n'):
        click.echo(line_format % line)


@click.command()
@click.argument('target')
@click.option('-f', '--force', is_flag=True, help="Run 'git push' with the '-f' force option")  # noqa
@click.option('-s', '--collectstatic', is_flag=True, help="Run collectstatic command post deployment")  # noqa
@click.option('-m', '--migrate', is_flag=True, help="Run the migrate command post deployment")  # noqa
@click.option('-b', '--branch', help="Deploy a specific branch")
@click.option('-v', '--verbose', is_flag=True, help="Show verbose output")
def deploy(target, force, collectstatic, migrate, branch, verbose):
    """Deploy app to target environment.

    Deploy a Heroku application and run post-deployment commands.

    Push code via git, run collectstatic and migrate commands, and wrap
    all of this with the maintenance page to prevent users from using the
    site whilst the deployment is running.

    The user is prompted to confirm various options prior to the
    deployment running, and is then required to enter a random number
    displayed on the screen to invoke the deployment.

    The function encacsulates a fixed workflow that maps git-flow to
    heroku environments, by pushing specific branches to specific remotes:

    """
    # return
    config = Configuration.load(DEFAULT_CONFIG)
    app_name = config.get_environment_setting(target, 'app_name')
    branch = branch or config.get_environment_setting(target, 'branch')
    cmd_collectstatic = config.get_command('collectstatic')
    cmd_migrate = config.get_command('migrate')

    remote_hash = get_remote_commit_hash(app_name)
    local_hash = get_local_commit_hash(branch)
    diff = git.get_diff(remote_hash, local_hash)
    log = git.get_log(remote_hash, local_hash)

    if verbose:
        click.echo("")
        click.echo("The following files have changed since the last deployment:\n")  # noqa
        split_print_lines(diff)
        click.echo("")
        click.echo("The following commits will be included in this deployment:\n")  # noqa
        split_print_lines(log)
        click.echo("")

    run_migrations = prompt_for_action(
        u"Do you want to run migrations?",
        ('/migrations/' in diff)
    )

    run_collectstatic = collectstatic or prompt_for_action(
        u"Do you want to run collectstatic?",
        ('/static/' in diff)
    )

    # summarise details
    click.echo("")
    click.echo("-------------------------------------")
    click.echo("Git branch:    %s" % branch)
    click.echo("Target env:    %s (%s)" % (target, app_name))
    click.echo("Force push:    %s" % force)
    # click.echo("Pipeline:            %s" % pipeline)
    # click.echo("Notify HipChat:      {hipchat}")
    click.echo("")
    click.echo("Post-deployment commands")
    click.echo("")
    click.echo("Data migrations: %s" % run_migrations)
    click.echo("Collect static:  %s" % run_collectstatic)
    click.echo("-------------------------------------")

    # release_note = get_release_note(log)

    # toggle_maintenance(True)

    # push_to_remote(force=force)

    # if run_migrate is True:
    #     migrate()

    # toggle_maintenance(False)

    # if run_collectstatic is True:
    #     collectstatic()



# -*- coding: utf-8 -*-
import os
import envoy

# location of the website git repo directory
WORK_TREE = os.getenv('WORK_TREE', ".")
GIT_DIR = "%s/.git" % WORK_TREE


def _do_git_cmd(command):
    """
    Run specified git command.

    This function is used to ensure that all git commands are run against
    the correct repository by including the the --git-dir and --work-tree
    options. This is required as we are running the commands outside of
    the target repo directory.

    Args:
        command: the command to run - without the 'git ' prefix, e.g.
            "status", "log --oneline", etc.

    Kwargs:
        capture: if True, then returns the output of the command as a string.

    Returns:
        the output of the command as a string if capture=True, else just
        runs the commands and returns None.
    """
    prefix = "git --git-dir=%s --work-tree=%s " % (GIT_DIR, WORK_TREE)
    r = envoy.run(prefix + command)
    if r.status_code > 0:
        raise Exception(
            u"Error running git command '%s': %s"
            % (command, r.std_err)
        )
    return r.std_out


def get_editor():
    """Return the configured editor value.

    Looks first in git.config, then $EDITOR, $VISUAL

    If no editor is found, returns None.

    """
    editor = (
        envoy.run('git config --get core.editor').std_out or
        os.getenv('EDITOR') or
        os.getenv('VISUAL')
    )
    if editor is None:
        raise Exception(
            "No editor configured in git config, "
            "$EDITOR or $VISUAL."
        )

    return editor


def checkout_branch(branch):
    """ Perform a 'git checkout' on the local repo."""
    _do_git_cmd("checkout %s" % branch)


def push(remote, local_branch, remote_branch="master", force=False):
    """Do a git push of a branch to a remote repo."""
    if force:
        _do_git_cmd("push %s %s:%s -f" % (remote, local_branch, remote_branch))
    else:
        _do_git_cmd("push %s %s:%s" % (remote, local_branch, remote_branch))


def get_branch_name():
    """ Return the current branch name (from GIT_DIR)."""
    return _do_git_cmd("rev-parse --abbrev-ref HEAD")


def get_log(commit_from, commit_to='HEAD'):
    """
    Return the oneline format git history between two commits.

    Args:
        commit_from: the commit hash of the earlier commit.

    Kwargs:
        commit_to: the commit hash of the later commit, defaults to HEAD.

    Returns:
        a list of 2-tuples, each containing the commit hash, and the commit
            message.

    e.g. if the commit log looks like this:

        81a5ea8 Fix for failing tests.
         Refactoring of conversations.

    The return value is:

        [
            ('81a5ea8', 'Fix for failing tests.'),
            ('62d49e9', Refactoring of conversations.)
        ]

    """
    command = "log --oneline --no-merges %s..%s" % (commit_from, commit_to)
    return _do_git_cmd(command)


def get_diff(commit_from, commit_to='HEAD'):
    """
    Return the names of all the files that have changed between two commits.

    Args:
        commit_from: the first commit - can be a commit hash or a tag.

    Kwargs:
        commit_to: the last commit - defaults to 'HEAD'. Can be a commit hash
            or a tag.

    Returns:
        a list of fully qualified filenames.
    """
    command = "diff --name-only %s..%s" % (commit_from, commit_to)
    return _do_git_cmd(command)
    # if len(diff) == 0:
    #     return None
    # else:
    #     return [d for d in diff.split('\n')]


def apply_tag(commit, tag, message):
    """Apply a tag to a given git commit."""
    command = "tag -a %s -m  '%s' %s" % (tag, message, commit)
    _do_git_cmd(command)



# -*- coding: utf-8 -*-
"""Heroku API bits and bobs."""
import os
import requests

# HEROKU_API_KEY = envoy.run("heroku auth:token", capture=True).std_out.rstrip()
HEROKU_API_KEY = os.getenv('HEROKU_API_TOKEN')
HEROKU_API_URL = 'https://api.heroku.com/apps/%s/releases'


def get_releases(application):
    """
    Get application release log from Heroku.

    Args:
        application: the name of the application on Heroku.

    Returns:
        JSON - see https://devcenter.heroku.com/articles/platform-api-reference#release
    """
    url = HEROKU_API_URL % application
    auth = requests.auth.HTTPBasicAuth('', HEROKU_API_KEY)
    return requests.get(url, auth=auth).json()


def get_latest_release(application):
    """
    Get the commit hash, version number of the latest release to Heroku.

    Args:
        application: the name of the application on Heroku.

    Returns:
        A 5-tuple containing:
            the version,
            commit hash,
            release description,
            user,
            date,
            e.g. (u'v31', u'7a0e825', u'Deployed', u'hugo', u'2013/07/11 04:46:37 -0700')
    """
    releases = get_releases(application)
    latest = releases[-1]
    return (
        latest["name"],
        latest["commit"],
        latest['descr'],
        latest['user'],
        latest['created_at']
    )
